IGNORE_INDEX = -100
DEFAULT_PAD_TOKEN = "[PAD]"
SPACE_TOKENIZERS = (
    "meta-llama/Meta-Llama-3-8B",
    "meta-llama/Meta-Llama-3-8B-Instruct",
    "meta-llama/Meta-Llama-3-70B",
    "meta-llama/Meta-Llama-3-70B-Instruct",
)
